/*Authors: Lou Davila Prado, Khalil El Ajeri
  Groupe: 14*/

#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>
#include <unistd.h>
#include <time.h>

#define CAMION_CAPACITE 4

typedef struct 
{
    sem_t velos;
    sem_t bornes;
    int velos_dispo;
    int personnes_attente;
    pthread_mutex_t mutex;
}Site;

int SITES,HABITANTS,TRAJETS,BORNES;
Site *sites;
pthread_mutex_t mutex_trajets = PTHREAD_MUTEX_INITIALIZER;
int trajets_effectues = 0;
int velos_depot = 0;
pthread_mutex_t mutex_velos_depot = PTHREAD_MUTEX_INITIALIZER;

void *habitant(void *arg) 
{
    int habitant_id = *(int *)arg;
    int site_actuel = rand()%SITES;
    free(arg);
    
    for (int i = 0; i < TRAJETS; i++) 
    {
        int site_suivant;
        do 
        {
            site_suivant = rand()%SITES;
        } 
        while (site_actuel == site_suivant);

        pthread_mutex_lock(&sites[site_actuel].mutex);
        sites[site_actuel].personnes_attente++;
        int velos = sites[site_actuel].velos_dispo;
        int attente = sites[site_actuel].personnes_attente;
        pthread_mutex_unlock(&sites[site_actuel].mutex);
        
        printf("(GET) Habitant %d attend un vélo au site %d (%d vélos, %d personnes en attente)\n", habitant_id, site_actuel, velos, attente);
        sem_wait(&sites[site_actuel].velos);
        
        pthread_mutex_lock(&sites[site_actuel].mutex);
        sites[site_actuel].velos_dispo--;
        sites[site_actuel].personnes_attente--;
        velos = sites[site_actuel].velos_dispo;
        attente = sites[site_actuel].personnes_attente;
        pthread_mutex_unlock(&sites[site_actuel].mutex);
        
        printf("(GET) Habitant %d prend un vélo au site %d (%d vélos, %d personnes en attente)\n", habitant_id, site_actuel, velos, attente);
        sem_post(&sites[site_actuel].bornes);
        //usleep((rand()%1000+1000));
    
        pthread_mutex_lock(&sites[site_suivant].mutex);
        sites[site_suivant].personnes_attente++;
        velos = sites[site_suivant].velos_dispo;
        attente = sites[site_suivant].personnes_attente;
        pthread_mutex_unlock(&sites[site_suivant].mutex);
        
        printf("(PUT) Habitant %d attend d'avoir accès à une borne du site %d (%d vélos, %d personnes en attente)\n", habitant_id, site_suivant, velos, attente);
        sem_wait(&sites[site_suivant].bornes);
        
        pthread_mutex_lock(&sites[site_suivant].mutex);
        sites[site_suivant].velos_dispo++;
        sites[site_suivant].personnes_attente--;
        velos = sites[site_suivant].velos_dispo;
        attente = sites[site_suivant].personnes_attente;
        pthread_mutex_unlock(&sites[site_suivant].mutex);
        
        sem_post(&sites[site_suivant].velos);
        printf("(PUT) Habitant %d dépose un vélo au site %d (%d vélos, %d personnes en attente)\n", habitant_id, site_suivant, velos, attente);
        //usleep((rand()%1000+1000));
        
        site_actuel = site_suivant;
        
        pthread_mutex_lock(&mutex_trajets);
        trajets_effectues++;
        pthread_mutex_unlock(&mutex_trajets);
    }
    return NULL;
}

void *gestion_camion(void *arg) 
{
    int velos_camion = 2;
    while (1) 
    {
        pthread_mutex_lock(&mutex_trajets);
        if (trajets_effectues >= HABITANTS*TRAJETS) 
        {
            pthread_mutex_unlock(&mutex_trajets);
            printf("Camion arrête son travail, tous les trajets sont effectués.\n");
            break;
        }
        pthread_mutex_unlock(&mutex_trajets);
        
        for (int i = 0; i < SITES; i++) 
        {
            pthread_mutex_lock(&sites[i].mutex);
            int velos = sites[i].velos_dispo;
            int attente = sites[i].personnes_attente;
            
            if (CAMION_CAPACITE > velos_camion && sites[i].velos_dispo > BORNES-2) 
            {
                int aRetirer = sites[i].velos_dispo-(BORNES-2);
                if (CAMION_CAPACITE-velos_camion < aRetirer)
                {
                    aRetirer = CAMION_CAPACITE-velos_camion;
                }
                
                if (sites[i].velos_dispo-1 < aRetirer)
                {
                    aRetirer = sites[i].velos_dispo-1;
                }
                
                pthread_mutex_unlock(&sites[i].mutex);
                
                for (int j = 0; j < aRetirer; j++) 
                {
                    sem_wait(&sites[i].velos);
                    
                    pthread_mutex_lock(&sites[i].mutex);
                    sites[i].velos_dispo--;
                    velos = sites[i].velos_dispo;
                    attente = sites[i].personnes_attente;
                    pthread_mutex_unlock(&sites[i].mutex);
                    
                    velos_camion++;
                    sem_post(&sites[i].bornes);
                }
                
                printf("Camion prend %d vélos au site %d (%d vélos, %d personnes en attente) (Vélos dans le camion: %d)\n", aRetirer, i, velos, attente, velos_camion);
            } 
            else if (sites[i].velos_dispo < 2 && velos_camion > 0) 
            {
                int aAjouter = 2-sites[i].velos_dispo;
                if (aAjouter > velos_camion)
                {
                    aAjouter = velos_camion;
                }
                
                int spotsDisponibles = BORNES-sites[i].velos_dispo;
                if (aAjouter > spotsDisponibles)
                {
                    aAjouter = spotsDisponibles;
                }
                
                pthread_mutex_unlock(&sites[i].mutex);
                
                for (int j = 0; j < aAjouter; j++) 
                {
                    sem_wait(&sites[i].bornes);
                    
                    pthread_mutex_lock(&sites[i].mutex);
                    sites[i].velos_dispo++;
                    velos = sites[i].velos_dispo;
                    attente = sites[i].personnes_attente;
                    pthread_mutex_unlock(&sites[i].mutex);
                    
                    velos_camion--;
                    sem_post(&sites[i].velos);
                }
                
                printf("Camion dépose %d vélos au site %d (%d vélos, %d personnes en attente) (Vélos dans le camion: %d)\n", aAjouter, i, velos, attente, velos_camion);
            }
            else 
            {
                pthread_mutex_unlock(&sites[i].mutex);
            }
            
            //usleep((rand()%100+100));
        }
        
        pthread_mutex_lock(&mutex_velos_depot);
        if (velos_camion > 2) 
        {
            velos_depot+= (velos_camion-2);
            velos_camion = 2;
            printf("Camion vide les vélos au dépôt (Vélos en dépôt: %d)\n", velos_depot);
        } 
        else if (velos_depot > 0 && velos_camion < 2) 
        {
            int requis = 2-velos_camion;
            if (requis > velos_depot)
            {
                requis = velos_depot;
            }
            velos_depot-= requis;
            velos_camion+= requis;
            printf("Camion prend des vélos au dépôt (Vélos en dépôt: %d)\n", velos_depot);
        }
        pthread_mutex_unlock(&mutex_velos_depot);
        
        printf("Camion fait un tour, Vélos restants: %d, vélos en dépôt: %d\n", velos_camion, velos_depot);
        //usleep((rand()%100+100));
    }
    return NULL;
}

int main(int argc, char *argv[]) 
{
    if (argc != 5) 
    {
        fprintf(stderr, "Usage: %s <SITES> <HABITANTS> <TRAJETS> <BORNES>\n", argv[0]);
        return 1;
    }
    SITES = atoi(argv[1]);
    HABITANTS = atoi(argv[2]);
    TRAJETS = atoi(argv[3]);
    BORNES = atoi(argv[4]);

    if (SITES <= 2 || HABITANTS <= 2 || TRAJETS <= 0 || BORNES <= 4)
    {
        printf("Les contraintes suivantes doivent être respectées:\n");
        printf("SITES > 2, HABITANTS > 2, TRAJETS > 0, BORNES > 4\n");
        exit(EXIT_FAILURE);
    }
    
    
    srand(time(NULL));
    sites = malloc(SITES * sizeof(Site));
    
    pthread_t habitants[HABITANTS], camion;
    
    for (int i = 0; i < SITES; i++) 
    {
        sem_init(&sites[i].velos, 0, BORNES-2);
        sem_init(&sites[i].bornes, 0, 2);
        pthread_mutex_init(&sites[i].mutex, NULL);
        sites[i].velos_dispo = BORNES-2;
        sites[i].personnes_attente = 0;
    }
    for (int i = 0; i < HABITANTS; i++) 
    {
        int *habitant_id = malloc(sizeof(int));
        *habitant_id = i;
        pthread_create(&habitants[i], NULL, habitant, habitant_id);
    }
    pthread_create(&camion, NULL, gestion_camion, NULL);
    
    for (int i = 0; i < HABITANTS; i++) 
    {
        pthread_join(habitants[i], NULL);
    }
    pthread_join(camion, NULL);
    
    int total_velos = velos_depot;
    for (int i = 0; i < SITES; i++) 
    {
        total_velos += sites[i].velos_dispo;
        printf("Site %d contient %d vélos\n", i, sites[i].velos_dispo);
    }
    total_velos += 2;
    printf("Nombre total de vélos à la fin : %d (doit être %d)\n", total_velos, SITES*(BORNES-2)+2);
    printf("Simulation terminée.\n");
    
    for (int i = 0; i < SITES; i++) 
    {
        sem_destroy(&sites[i].velos);
        sem_destroy(&sites[i].bornes);
        pthread_mutex_destroy(&sites[i].mutex);
    }
    pthread_mutex_destroy(&mutex_trajets);
    pthread_mutex_destroy(&mutex_velos_depot);
    free(sites);
    return 0;
}